package com.urbancode.anthill3.plugin.rtcscm

import com.urbancode.anthill3.CommandHelper
import com.urbancode.anthill3.AHPTool
import java.text.SimpleDateFormat

public class RTCHelper {

	final def command
    final def dirOffset
    final def url
    final def username
	final def password
	final def workspace
	final def stream
	final def baseline
	final def snapshot
	final def components
    final def anthillSnapshot
    final def anthillSnapshotDescription
    final def anthillSnapshotProperty
    final def isIncludeRoot
    final def isForce
    final def listSnapshotDateFormat
    final static def CHANGELOG_DATE_FORMAT = 'yyyy-MM-dd HH:mm:ss.S Z'

    final def out = System.out
    final def workDir = new File('.').canonicalFile
    final def ahpTool = new AHPTool();


	public RTCHelper(Map<String,String> props) {
		def commandPath = props['source/repo/commandPath']
        dirOffset = props['source/dirOffset']
        url = props['source/repo/url']
        username = props['source/repo/username']
        password = props['source/repo/password']
        workspace = props['source/workspace']
        stream = props['source/stream']
        baseline = props['source/baseline']
        snapshot = props['source/snapshot']
        components = props['source/components']
        anthillSnapshot = props['source/anthillSnapshot']
        anthillSnapshotDescription = props['source/anthillSnapshotDescription']?:' '
        anthillSnapshotProperty = 'rtc.snapshot'
        isIncludeRoot = (props['source/isIncludeRoot']?:'false').toBoolean()
        isForce = (props['source/isForce']?:'false').toBoolean()
        listSnapshotDateFormat = props['source/repo/listSnapshotDateFormat']


        command = (commandPath ?: 'scm').replace('/', System.getProperty('file.separator')).replace('\\', System.getProperty('file.separator'))

	}

    public void cleanup(dir) {
        if (dir == null) {
            dir = new File(workDir, dirOffset ?: '.').canonicalFile
        }
        println "Removing everything in ${dir}..."
        dir.listFiles().each {file ->
            boolean success = false;
            if (file.isFile()) {
                success = file.delete()
            }
            else if (file.isDirectory()) {
                success = file.deleteDir()
            }
            if (!success) {
                println("Failed to delete $file")
            }
        }
        println "Done deleteing directory contents for ${dir}"
    }

    public void undo(dir) {
        if (dir == null) {
            dir = new File(workDir, dirOffset ?: '.').canonicalFile
        }
        def statusText
        getCmdHelper(dir).runCommand('Get workspace status', [command, '--non-interactive', 'status', '-u', username, '-P', password]) {proc ->
            proc.consumeProcessErrorStream(out)
            proc.outputStream.close()
            statusText = proc.text.trim()
            println statusText
        }
        def isChange = false
        def isWorkspaceMatch = false
        def undoCommand = []
        undoCommand << command << '--non-interactive' << 'undo' << '-u' << username << '-P' << password
        statusText.eachLine {it ->
            def line = it.trim()
            if (line.startsWith('Workspace')) {
                line.find('Workspace: \\((\\d+)\\) "(.*)" <->.*') {match, workspaceAlias, workspaceName ->
                    if (workspaceAlias.equalsIgnoreCase(workspace) || workspaceName.equalsIgnoreCase(workspace)) {
                        isWorkspaceMatch = true
                    }
                    else {
                        println "Local workspace ${workspaceAlias}:${workspaceName} does not match required remote workpsace ${workspace}!"
                    }
                }
            }
            else if (line.startsWith('Unresolved')) {
                isChange = true
            }
            else if (isChange && (line.startsWith('a-') || line.startsWith('d-') || line.startsWith('m-') || line.startsWith('-c'))) {
                undoCommand << line.substring(4)
            }
            else if (isChange) {
                isChange = false
            }
        }

        //now run the undo command on all local changes unless the workspaces don't match
        if (!isWorkspaceMatch) {
            cleanup(dir)
        }
        else if (undoCommand.size() > 7) {
            getCmdHelper(dir).runCommand('Undo local changes', undoCommand)
        }
        else {
            println 'No local changes found, nothing to cleanup!'
        }
    }

    public Map getWorkspaceInfo(dir) {
        if (dir == null) {
            dir = workDir
        }
        def componentText
        def error = new StringBuffer()
        def result = [:]
        result.componentList = [:]
        try {
            getCmdHelper(dir).runCommand('Get workspace components', [command, '--non-interactive', 'list', 'components', workspace, '-u', username, '-P', password, '-r', url]) {proc ->
                proc.consumeProcessErrorStream(error)
                proc.outputStream.close()
                componentText = proc.text.trim()
                println componentText
                println error
                result.cmdError = error
            }
        }
        catch (Exception e) {}
        if (error.length() == 0) {
            componentText.eachLine {it ->
                def line = it.trim()
                if (line.startsWith('Workspace')) {
                    line.find('.*<-> \\((\\d+)\\) "(.*)"') {match, streamAlias, streamName ->
                        result.workspaceStreamAlias = streamAlias
                        result.workspaceStreamName = streamName
                    }
                } else if (line.startsWith('Component') && line[-1] != ')') {
                    line.find('Component: \\((\\d+)\\) "(.+)"') {match, compAlias, compName ->
                        result.componentList."${compAlias}" = compName
                    }
                }
            }
        }        
        return result
    }

    public Map getStreamOrSnapshotComponents() {
        def componentText
        def result = [:]
        def cmdArgs = [command, '--non-interactive', 'list', 'components'];
        if (snapshot) {
            cmdArgs << '-s';
        }
        cmdArgs << '-u';
        cmdArgs << username;
        cmdArgs << '-P';
        cmdArgs << password;
        cmdArgs << '-r';
        cmdArgs << url;

        if (snapshot) {
            cmdArgs << snapshot;
        }
        else {
            cmdArgs << stream;
        }
        getCmdHelper(workDir).runCommand('List stream components', cmdArgs) {proc ->
            proc.consumeProcessErrorStream(out)
            proc.outputStream.close()
            componentText = proc.text.trim()
            println componentText
        }
        componentText.eachLine {it ->
            def line = it.trim()
            if (line.startsWith('Component') && line[-1] != ')') {
                line.find('Component: \\((\\d+)\\) "(.+)"') {match, compAlias, compName ->
                    result."${compAlias}" = compName
                }
            }
        }
        return result
    }

    public def getStreamSnapshots() {
        def snapshotText
        def result = []
        def dateFormat = new SimpleDateFormat(listSnapshotDateFormat)
        getCmdHelper(workDir).runCommand('List stream snapshots', [command, '--non-interactive', 'list', 'snapshots',
                stream,
                '-u', username,
                '-P', password,
                '-r', url]) {proc ->
            proc.consumeProcessErrorStream(out)
            proc.outputStream.close()
            snapshotText = proc.text.trim()
            println snapshotText
        }
        snapshotText.eachLine {it ->
            def line = it.trim()
            line.find('\\((\\d+)\\) "(.+)" (.*)') {match, alias, name, date ->
                result << [alias:alias, date:dateFormat.parse(date)]
            }
        }
        return result
    }

    public void createSnapshot() {
        def currentBuildlifeSnapshot = ahpTool.getBuildLifeProperties()["${anthillSnapshotProperty}"]
        if (currentBuildlifeSnapshot) {
            println "Snapshot has already been created, nothing to do!"
        }
        else {
            if (snapshot || baseline) {
                println "Building from ${snapshot?"snapshot":"baseline"} so no need to create a new snapshot!"
            }
            else {
                getCmdHelper(workDir).runCommand('Create snapshot', [command, '--non-interactive', 'create', 'snapshot',
                        '-n', anthillSnapshot,
                        '-d', anthillSnapshotDescription,
                        stream,
                        '-r', url,
                        '-u', username,
                        '-P', password])
            }
            def newProps = ["${anthillSnapshotProperty}":"${snapshot?:(baseline?:anthillSnapshot)}"]
            if (baseline) {
                newProps['rtc.isBaseline'] = 'true'
            }
            ahpTool.setBuildLifeProperties(newProps)
        }
    }

    public void createWorkspace() {
        def cmd = [command, '--non-interactive', 'create', 'workspace',
                '-d', 'Created by AnthillPro',
                '-r', url,
                '-u', username,
                '-P', password]
        if (snapshot) {
            cmd << '--snapshot' << snapshot
        }
        else if (baseline) {
            cmd << '-s' << stream
        }
        else {
            cmd << '--snapshot' << anthillSnapshot
        }
        cmd << workspace

        getCmdHelper(workDir).runCommand('Create workspace', cmd)
    }

    public void updateWorkspaceComponents(currentComponentList) {

        if (currentComponentList) {
            currentComponentList.each { key, value ->
                currentComponentList."${key}" = currentComponentList."${key}".toLowerCase()
            }
            def removeComponentList = []
            def addComponentList = []

            if (components) {
                // user provided a list of components
                components.split(',').each {it ->
                    if (!currentComponentList.containsKey(it.toLowerCase()) && !currentComponentList.containsValue(it.toLowerCase())) {
                        addComponentList << it
                    }
                }
                def componentList = components.toLowerCase().split(',').toList()
                currentComponentList.each { key, value ->
                    if (!componentList.contains(key) && !componentList.contains(value)) {
                        removeComponentList << key
                    }
                }
            }
            else {
                // user did not provide component list, assuming all
                def targetList = getStreamOrSnapshotComponents()
                targetList.each { key, value ->
                    if (!currentComponentList.containsKey(key)) {
                        addComponentList << key
                    }
                }
                currentComponentList.each { key, value ->
                    if (!targetList.containsKey(key)) {
                        removeComponentList << key
                    }
                }
            }

            if (removeComponentList.size() > 0) {
                removeComponents(removeComponentList)
            }
            if (addComponentList.size() > 0) {
                addComponents(addComponentList)
            }
        }
    }

    private void addComponents(componentList) {
        def cmd = [command, '--non-interactive', 'workspace', 'add-components',
                '-r', url,
                '-u', username,
                '-P', password,
                workspace]
        componentList.each {it ->
            cmd << it
        }
        getCmdHelper(workDir).runCommand('Add Workspace components', cmd)
    }

    private void removeComponents(componentList) {
        def cmd = [command, '--non-interactive', 'workspace', 'remove-components',
                '-r', url,
                '-u', username,
                '-P', password,
                workspace]
        componentList.each {it ->
            cmd << it
        }
        getCmdHelper(workDir).runCommand('Remove Workspace components', cmd)
    }

    public void changeWorkspaceBaseline() {
        if (baseline) {
            getCmdHelper(new File(workDir, '..').canonicalFile).runCommand('Change Workspace Baseline',
                    [command, '--non-interactive', 'workspace', 'rplc',
                    '-o',
                    '-b', baseline,
                    '--all',
                    '-r', url,
                    '-u', username,
                    '-P', password,
                    workspace,
                    'stream',
                    stream])
        }
    }

    public void changeWorkspaceSnapshot() {
        getCmdHelper(new File(workDir, '..').canonicalFile).runCommand('Change Workspace Snapshot',
                [command, '--non-interactive', 'workspace', 'rplc',
                '-o',
                '--all',
                '-r', url,
                '-u', username,
                '-P', password,
                workspace,
                'snapshot',
                snapshot?:anthillSnapshot])
    }

    public void changeWorkspaceFlowStream(currentStream, currentStreamAlias) {
        if (stream && !stream.equalsIgnoreCase(currentStream) && !stream.equalsIgnoreCase(currentStreamAlias)) {
            getCmdHelper(new File(workDir, '..').canonicalFile).runCommand('Change Workspace Flow Stream',
                    [command, '--non-interactive', 'change-target', 'workspace',
                    '-r', url,
                    '-u', username,
                    '-P', password,
                    workspace,
                    stream])
        }
    }

    public void loadWorkspace() {
        def dir = new File(workDir, dirOffset?:'.').canonicalFile
        if (!dir.exists()) {
            dir.mkdirs()
        }
        def cmd = [command, '--non-interactive', 'load',
                '--all',
                '-d', dir.toString(),
                '-r', url,
                '-u', username,
                '-P', password]
        if (isForce) {
            cmd << '-f'
        }
        if (isIncludeRoot) {
            cmd << '-i'
        }
        cmd << workspace
        getCmdHelper(dir).runCommand('Load workspace', cmd)
    }

    public void repair() {
        getCmdHelper(new File(workDir, dirOffset ?: '.').canonicalFile).runCommand('Attempt local workspace repair',
                [command, '--non-interactive', 'repair',
                '-r', url,
                '-u', username,
                '-P', password])
    }

    /**
     * (1032) User:anthill Comment: 7, 15: Test work item - Change readme file.And try a new line Date:2010-03-16 10:13:53 -0500
     *      Work Item 7: Test - work - item
     *      Work Item 15: Test task
     *      /conf/jalopy.xml
     *      /readme.txt
     *
     */
    public def getChangelog(start, startType, end, endType, componentList) {
        def changeSetList = []
        if (start && end) {
            if (componentList && componentList.size() > 0) {
                componentList.each { component ->
                    def changelogText
                    getCmdHelper(workDir).runCommand('Get Changelog',
                            [command, '--non-interactive', '--show-uuid', 'y', 'compare',
                                    startType, start,
                                    endType, end,
                                    '-I', 'wsf',
                                    '-f', 'i',
                                    '-C', 'User:{userid} Comment:',
                                    '-p', 'd',
                                    '-S', 'cdi',
                                    '-D', "'Date:'" + CHANGELOG_DATE_FORMAT,
                                    '-c', component,
                                    '-u', username,
                                    '-P', password,
                                    '-r', url]) {proc ->
                        proc.consumeProcessErrorStream(out)
                        proc.outputStream.close()
                        changelogText = proc.text.trim()
                        println changelogText
                        changeSetList += parseChangelog(changelogText)
                    }
                }
            }
            else {
                def changelogText
                getCmdHelper(workDir).runCommand('Get Changelog',
                        [command, '--non-interactive', '--show-uuid', 'y', 'compare',
                                startType, start,
                                endType, end,
                                '-I', 'wsf',
                                '-f', 'i',
                                '-C', 'User:{userid} Comment:',
                                '-p', 'd',
                                '-S', 'cdi',
                                '-D', "'Date:'" + CHANGELOG_DATE_FORMAT,
                                '-u', username,
                                '-P', password,
                                '-r', url]) {proc ->
                    proc.consumeProcessErrorStream(out)
                    proc.outputStream.close()
                    changelogText = proc.text.trim()
                    println changelogText
                    changeSetList += parseChangelog(changelogText)
                }
            }
        }
        return changeSetList
    }

    /**
     * Converts an Ant-style wild-mat string into an equivalent java.util.regex.Pattern
     *   ? is a single char wildcard
     *   * is a wildcard matches any string (but does not cross directories)
     *   ** is a wildcard which matches any string and can cross any number of directories
     */
    public def convertToPattern(def antWildMat) {
        // trim leading / character from pattern
        def pattern = antWildMat.startsWith('/') ? antWildMat : '/' + antWildMat

        // deal with special regex-characters that should be interpreted as literals
        '\\.+[]^${}|()'.toCharArray().each {c ->
            pattern = pattern.replace('' + c, '\\' + c)
        }
        pattern = pattern.replace('?', '.') // ? is a single-char wildcard

        // deal with ant-style wildcards
        StringBuffer result = new StringBuffer()
        def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
        while (m) {
            def token = m.group()
            def replacement;
            if (token == '**/') {
                replacement = '.*(?<=/)'
            } else if (token == '**') {
                replacement = '.*'
            } else {
                replacement = '[^/]*'
            }
            m.appendReplacement(result, java.util.regex.Matcher.quoteReplacement(replacement))
        }
        m.appendTail(result)
        return ~(result.toString())
    }

    private def parseChangelog(changelogText) {
        def changelist = []
        def currentChangeSet
        def currentWorkItemList = []
        changelogText.eachLine { it ->
            def line = it.trim()
            // 3.0
            if (line ==~ '\\(\\d+\\) User:.*') {
                line.find('\\((\\d+)\\) User:(.*) Comment:(.*) Date:(.*)') {match, id, user, comment, date ->
                    if (currentChangeSet) {
                        if (currentWorkItemList.size() > 0) {
                            currentChangeSet.message = currentChangeSet.message.substring(currentChangeSet.message.indexOf(':'))
                            currentChangeSet.message = currentChangeSet.message.substring(currentWorkItemList[0].name.size() + 3)
                        }
                        changelist << currentChangeSet
                        currentWorkItemList.clear()
                    }
                    currentChangeSet = [id:id, user:user, message:comment, date:date, fileSet:[], issueId:'']
                }
            }
            else if (line.startsWith('Work Item')) {
                line.find('Work Item (\\d+): (.+)') {match, workItemId, name ->
                    currentWorkItemList << [id:workItemId, name:name]
                    if (currentChangeSet.issueId.size() > 0) {
                        currentChangeSet.issueId += ','
                    }
                    currentChangeSet.issueId += workItemId
                }
            }
            // 4.0
            else if (line ==~ ('\\(\\d+:_[0-9A-Za-z\\-_]+\\) User:.*')) {
                line.find('\\(\\d+:(_[0-9A-Za-z\\-_]+)\\) User:(.*) Comment:(.*) Date:(.*)') {match, id, user, comment, date ->
                    if (currentChangeSet) {
                        if (currentWorkItemList.size() > 0) {
                            currentChangeSet.message = currentChangeSet.message.substring(currentChangeSet.message.indexOf(':'))
                            currentChangeSet.message = currentChangeSet.message.substring(currentWorkItemList[0].name.size() + 3)
                        }
                        changelist << currentChangeSet
                        currentWorkItemList.clear()
                    }
                    currentChangeSet = [id:id, user:user, message:comment, date:date, fileSet:[], issueId:'']
                }
            }
            else if (line ==~ ('\\(\\d+:_[0-9A-Za-z\\-_]+\\) Work Item.*')) {
                line.find('Work Item (\\d+): (.+)') {match, workItemId, name ->
                currentWorkItemList << [id:workItemId, name:name]
                        if (currentChangeSet.issueId.size() > 0) {
                            currentChangeSet.issueId += ','
                        }
                currentChangeSet.issueId += workItemId
                }
            }
            else if (line.size() > 0){
                assert currentChangeSet:'Unrecognized line in changeset!'
                currentChangeSet.fileSet << line
            }
        }
        if (currentChangeSet) {
            changelist << currentChangeSet
            if (currentWorkItemList.size() > 0) {
                currentChangeSet.message = currentChangeSet.message.substring(currentChangeSet.message.indexOf(':') + 1)
                // currentChangeSet.message = currentChangeSet.message.substring(currentWorkItemList[0].name.size() + 4)
            }
        }

        return changelist
    }

    private CommandHelper getCmdHelper(dir) {
        return new CommandHelper(dir)
    }
}
