import com.urbancode.anthill3.AHPTool
import com.urbancode.anthill3.plugin.rtcscm.RTCHelper

final def workDir = new File('.').canonicalFile
final def ahpTool = new AHPTool();
final def props = ahpTool.getStepProperties();
final def userExcludes = props['source/excludeUsers']?.readLines() ?: []
final def fileFilters = props['source/fileFilters']?.readLines() ?: []
final def components = props['source/components']
final def persistToDB = props['persistingChanges']?.toBoolean()
//final def startBuildlifeId = props['startBuildlifeId']?.toLong()
final def startBuildLifeId = (props['startBuildlifeId'] != null && props['startBuildlifeId'] != "") ?
		props['startBuildlifeId'].toLong() : null;
def startSnapshot
def isStartBaseline
try {
    startSnapshot = ahpTool.getBuildLifeProperties(startBuildLifeId)['rtc.snapshot']
    println 'Starting build life ID: ' + startBuildLifeId
    println 'Starting snapshot: ' + startSnapshot
    isStartBaseline = ahpTool.getBuildLifeProperties(startBuildLifeId)['rtc.isBaseline']?.toBoolean()
}
catch (Exception e) {
	//swallow exception to allow not having an initial buildLife without failing the step
	println(e.getMessage());
}

final def endSnapshot = ahpTool.getBuildLifeProperties()['rtc.snapshot']
final def isEndBaseline = ahpTool.getBuildLifeProperties()['rtc.isBaseline']?.toBoolean()
final def rtcHelper = new RTCHelper(props)

if (userExcludes) {
  userExcludes = userExcludes.collect{it.trim()}.findAll{it}
}

def includeFiles = []
def excludeFiles = []
if (fileFilters) {
  fileFilters.collect{it.trim()}.findAll{it}.each{filter ->
      def pattern = rtcHelper.convertToPattern(filter.substring(1))
      if (filter.startsWith('+')) {
          includeFiles << pattern
      }
      else if (filter.startsWith('-')) {
          excludeFiles << pattern
      }
      else {
          // not a valid expression
      }
  }
}

/**
 * determine if the given path is a non-excluded path belonging to the desired set
 *
 * @param path the file path to be tested
 */
def pathIsIncluded = { path ->
    boolean included = false;

    for (include in includeFiles) {
        if (path ==~ include) {
            included = true
            break
        }
    }

    for (exclude in excludeFiles) {
        if (path ==~ exclude) {
            included = false
            break
        }
    }
    return included;
}

def publishChangesets(changeSets, ahpTool) {
    def xml = groovy.xml.DOMBuilder.newInstance().'change-log' {
        changeSets.each {changeSet ->
            'change-set'() {
                id(changeSet.id)
                user(changeSet.user)
                date(changeSet.date)
                if (changeSet.message) {
                    comment(changeSet.message)
                }
                'repository-type'('RTC')
                'repository-id'(System.getenv()['AH_REPOSITORY_ID'])
                'file-set' {
                    changeSet.fileSet.each {change ->
                        file(change)
                    }
                }
                'properties' {
                    'property' {
                        name('rtc.workitems')
                        value(changeSet.issueId)
                    }
                }
            }
        }
    }

    ahpTool.addChangeSets(xml)
}

if (startSnapshot && endSnapshot) {
    def componentSet
    if (!components && (isStartBaseline || isEndBaseline)) {
        componentSet = rtcHelper.getWorkspaceInfo(null).componentList.keySet()
    }
    else if (components) {
        componentSet = components.split(',')
    }
    def changeSets = rtcHelper.getChangelog(startSnapshot, isStartBaseline ? 'baseline' : 'snapshot',
            endSnapshot, isEndBaseline ? 'baseline' : 'snapshot', componentSet)
    def filteredChangeSets = []
    changeSets.each{ change ->
        if (!userExcludes || !userExcludes.contains(change.user)) {
            def include = false
            if (fileFilters) {
                def fileSet = change.fileSet
                fileSet.each{ path ->
                    if (pathIsIncluded(path)) {
                        include = true
                    }
                    else {
                        change.fileSet -= path
                    }
                }
                if (include) {
                    filteredChangeSets << change
                }
            }
            else {
                filteredChangeSets << change
            }
        }
    }
    println "Filtered Change Sets: " + filteredChangeSets
    if (persistToDB) {
        publishChangesets(filteredChangeSets, ahpTool)
    }

}
else {
    println "Start Buildlife ID: ${startBuildLifeId}"
    println "Start Snapshot: ${startSnapshot} ${isStartBaseline?'is basline':''}"
    println "End Snapshot: ${endSnapshot} ${isEndBaseline?'is basline':''}"
    println "Cannot get a changelog without start and end snapshots!"
}





