import com.urbancode.anthill3.AHPTool
import com.urbancode.anthill3.plugin.rtcscm.RTCHelper
import java.text.SimpleDateFormat

final def workDir = new File('.').canonicalFile
final def ahpTool = new AHPTool();
final def props = ahpTool.getStepProperties();
final def userExcludes = props['source/excludeUsers']?.readLines() ?: []
final def fileFilters = props['source/fileFilters']?.readLines() ?: []
final def components = props['source/components']
final def stream = props['source/stream']
final def baseline = props['source/baseline']
final def snapshot = props['source/snapshot']
final def startTime = props['startDate']?.toLong()
final def rtcHelper = new RTCHelper(props)

println "Start Date: " + startTime

if (userExcludes) {
  userExcludes = userExcludes.collect{it.trim()}.findAll{it}
}

def includeFiles = []
def excludeFiles = []
if (fileFilters) {
  fileFilters.collect{it.trim()}.findAll{it}.each{filter ->
      def pattern = rtcHelper.convertToPattern(filter.substring(1))
      if (filter.startsWith('+')) {
          includeFiles << pattern
      }
      else if (filter.startsWith('-')) {
          excludeFiles << pattern
      }
      else {
          // not a valid expression
      }
  }
}

/**
 * determine if the given path is a non-excluded path belonging to the desired set
 *
 * @param path the file path to be tested
 */
def pathIsIncluded = { path ->
    boolean included = false;

    for (include in includeFiles) {
        if (path ==~ include) {
            included = true
            break
        }
    }

    for (exclude in excludeFiles) {
        if (path ==~ exclude) {
            included = false
            break
        }
    }
    return included;
}

if (stream && !snapshot && !baseline) {
    def startSnapshot
    def snapshotList = rtcHelper.getStreamSnapshots()
    for (i in 0..snapshotList.size()) {
        if (snapshotList[i].date.getTime() < startTime) {
            startSnapshot = snapshotList[i].alias
            break
        }
    }
    if (!startSnapshot) {
        println "Could not locate a previous snapshot on stream ${stream} that was created before ${new Date(startTime)}."
        println "Assuming there are no changes!"
    }
    else {
        def componentSet
        if (components) {
            componentSet = components.split(',')
        }
        def changeSets = rtcHelper.getChangelog(startSnapshot, 'snapshot', stream, 'stream', componentSet)
        def filteredChangeSets = []
        changeSets.each{ change ->
            if (!userExcludes || !userExcludes.contains(change.user)) {
                def include = false
                if (fileFilters) {
                    def fileSet = change.fileSet
                    fileSet.each{ path ->
                        if (pathIsIncluded(path)) {
                            include = true
                        }
                        else {
                            change.fileSet -= path
                        }
                    }
                    if (include) {
                        filteredChangeSets << change
                    }
                }
                else {
                    filteredChangeSets << change
                }
            }
        }
        println "Filtered Change Sets: " + filteredChangeSets
        def latestDate
        def dateFormat = new SimpleDateFormat(RTCHelper.CHANGELOG_DATE_FORMAT)
        filteredChangeSets.each{ change ->
            def time = dateFormat.parse(change.date)?.getTime()
            if (time > latestDate && time > startTime) latestDate = time
        }
        if (latestDate) {
            ahpTool.setQuietPeriodLatestTime(latestDate)
        }
    }
}
else if (snapshot || baseline) {
    println "Building from snapshot or baseline so no need for quiet period!"
}
else {
    println "No stream specified!"
}




